#!/usr/bin/env python
# -*- coding: utf-8 -*-

from re import compile as re_compile
import sqlite3
from traceback import print_exc as print_trace_exception
from sys import exc_info as exception_information
import traceback
import json

class Anonymous:

    anp = None
    __closed = True
    __errors = 0
    __name = "default"
    __autotransaction = True
    __input = None
    __temporary_table_header = ""
    __order_translate = {
        "ascendent" : "asc",
        "descendent" : "desc"
    }

    __re_string_escapes = re_compile(r'([\'])')
    __re_query = re_compile(r'(([^\';\\\\]+|\\\\(.|[\r\n])|\'(([^\'\\\\]+|\\\\(.|[\r\n]))*)\')+)')

    def __init__(self, anp, input = None):
        self.anp = anp
        self.anp._print("info", "sqlite_building", {
            "name" : self.__name,
            "error" : self.__errors
        })

        self.__input = input
        self.__temporary_table_header = self.anp.settings.get(["sqlite_temporary_table_header", "temporary_table_header"], self.__input)

        if not input:
            self.__errors |= 1 << 3
            self.anp._print("warn", "sqlite_no_input", {
                "name" : self.__name,
                "error" : self.__errors
            })
            return

        if AnP.is_string(input):
            input = {"file" : input}

        if "name" in input and input["name"]:
            self.__name = input["name"]

        if "file" not in input:
            self.__errors |= 1 << 4
            self.anp._print("warn", "sqlite_no_file", {
                "name" : self.__name,
                "error" : self.__errors
            })
            return

    @classmethod
    def escape(self, value):
        if AnP.is_string(value):
            return self.__re_string_escapes.sub('\'\'', value)
        return value

    def process_variables(self, variables = None):
        if not variables:
            return None
        if not AnP.is_array(variables):
            variables = [variables]
        for i in range(len(variables)):
            if AnP.is_dictionary(variables[i]):
                for key, value in variables[i].items():
                    if AnP.is_array(value) or AnP.is_dictionary(value):
                        json = self.escape(self.anp.json_encode(value))
                        variables[i][key] = "'" + json + "'" if json else "null"
                        continue
                    variables[i][key] = (
                        "null" if value == None else
                        "'" + self.escape(value) + "'" if AnP.is_string(value) else
                        ("true" if value else "false") if AnP.is_bool(value) else
                        "'" + value.strftime("%Y-%d-%m %H:%M:%S") + "'" if AnP.is_datetime(value) else
                        value
                    )
        return variables

    @classmethod
    def multi_queries_array(self, origin):

        queries = []

        for sql in [] if not origin else origin if AnP.is_array(origin) else [origin]:
            for query in self.__re_query.findall(sql):
                queries.append(query[0])

        return queries

    def connect(self):

        try:

            file = self.anp.format_path(self.__input["file"])

            try:
                AnP.make_directories(AnP.get_directory_path(file))
            except:
                self.anp._print("error", "unaccessible_db_unit")
                return None

            connection = sqlite3.connect(file)
            connection.row_factory = sqlite3.Row

            return connection
        except:
            print_trace_exception()
            pass

        return None

    def close(self, connection):
        try:
            connection.close()
            return True
        except:
            pass
        return False

    def query(self, queries, variables = None, connection = None):

        results = {
            "data" : [],
            "errors" : self.__errors
        }

        if self.__errors != 0:
            results["errors"] |= 1 << 1
            return results

        try:

            autoclose = connection == None

            if autoclose:
                connection = self.connect()
            if not connection:
                results["errors"] |= 1 << 5
                return results

            cursor = connection.cursor()

            variables = self.process_variables(variables.copy()) if AnP.is_dictionary(variables) or AnP.is_array(variables) else None

            for query in self.multi_queries_array(queries):

                id = None
                error = 0
                data = []

                try:
                    subquery = AnP.string_variables(query, variables)
                    # print(subquery)
                    cursor.execute(subquery)
                    id = cursor.lastrowid
                except:
                    print({"subquery" : subquery})
                    print_trace_exception()
                    error |= 1 << 0

                try:
                    data = cursor.fetchall()
                    data = [dict(row) for row in data] if data else []
                except:
                    data = []
                    pass

                results["data"].append({
                    "table" : data,
                    "error" : error,
                    "id" : id
                })

            connection.commit()

        except:
            print({"queries" : queries})
            print_trace_exception()
            # No guardar la excepción por el hecho de la recursividad.
            connection.rollback()
        finally:
            if autoclose:
                self.close(connection)

        return results

    @staticmethod
    def type(self):
        return "sqlite"

    def execute_sql(self, url, variables = None):

        sql = self.anp.file_load(url)

        return self.query(sql.decode("utf-8") if AnP.python3() else sql, variables)

    @staticmethod
    def datetime_format(date):
        if AnP.is_number(date):
            date = AnP.datetime(date / 1000)
        return date.strftime("%Y-%m-%d %H:%M:%S")

    def list(self, view, page, items_per_page, where, order_by, error = 0, at_first = None, variables = None):

        table_name = None
        order = ""
        search = ""

        if order_by and len(order_by):
            for field in order_by:
                order += ", " + field[0] + " " + ("desc" if field[1] else "asc")
            order = " order by" + order[1:]

        if not where:
            where = [None]

        if AnP.is_array(at_first) and len(at_first):
            for item in at_first if AnP.is_array(at_first[0]) else [at_first]:
                where.insert(1, item)
        if AnP.is_array(variables) and len(variables):
            for item in variables if AnP.is_array(variables[0]) else [variables]:
                where.append(item)

        if where[0]:

            temporary_view = self.__temporary_table_header + self.anp.hash()

            for column in self.query((
                "create table " + temporary_view + " as select * from " + view + ";" +
                "select * from (select name, (case " +
                	"when \"type\" in ('integer', 'real', 'numeric', 'double', 'float', 'decimal') or " +
                		"\"type\" like '%int%' or " +
                		"\"type\" like 'decimal%' or " +
                		"\"type\" like 'numeric%' or " +
                		"\"type\" like 'double%' " +
                	"then 'number' " +
                	"when \"type\" in ('varchar', 'text', 'char', 'nchar', 'nvarchar') or " +
                		"\"type\" like '%char%' or " +
                		"\"type\" like '%text%' " +
                	"then 'text' " +
                "else 'ignore' end) as \"type\" from pragma_table_info({temporary_view})) where \"type\" != 'ignore';" +
                "drop table " + temporary_view + ";"
            ), {"temporary_view" : temporary_view})["data"][1]["table"]:
                search += " or " + ("\"" + column["name"] + "\"" if column["type"] == "text" else "(\"" + column["name"] + "\" || '')") + " like '%" + where[0] + "%'"
            if search:
                search = "(" + search[3:] + ")"

            where = where[1:]

        if len(where):
            filter = ""
            for field in where:
                if not field:
                    continue
                filter += " and "
                if field[1] == None:
                    filter += "\"" + field[0] + "\" is" + (" not" if field[2] == True else "") + " null"
                elif field[2] in ["string", "text", "varchar"]:
                    filter += "\"" + field[0] + "\" like '%" + field[1] + "%'"
                elif field[2] in ["integer", "number"]:
                    filter += "\"" + field[0] + "\""
                    if AnP.is_array(field[1]):
                        if field[1][0] == None:
                            filter += " <= " + str(field[1][1])
                        elif field[1][1] == None:
                            filter += " >= " + str(field[1][0])
                        else:
                            filter += " between " + str(field[1][0]) + " and " + str(field[1][1])
                    else:
                        filter += " = " + str(field[1])
                elif field[2] in ["datetime"]:
                    filter += "\"" + field[0] + "\""
                    if AnP.is_array(field[1]):
                        if field[1][0] == None:
                            filter += " <= '" + self.datetime_format(field[1][1]) + "'"
                        elif field[1][1] == None:
                            filter += " >= '" + self.datetime_format(field[1][0]) + "'"
                        else:
                            filter += " between '" + self.datetime_format(field[1][0]) + "' and '" + self.datetime_format(field[1][1]) + "'"
                    else:
                        filter += " = " + self.datetime_format(field[1])
                elif field[2] in ["bool", "boolean"]:
                    filter += ("" if field[1] else "not ") + "\"" + field[0] + "\""
            if filter:
                search += filter if search else filter[4:]

        search = " where " + ("false" if error else "true") + (" and" + search if search else "")

        while True:
            table_name = self.__temporary_table_header + self.anp.hash()
            if not len(self.query("select 1 from sqlite_master where type = 'table' and name = {name}", {"name" : table_name})["data"][0]["table"]):
                break

        items = 0

        try:
            items = self.query((
                "create table " + table_name + " as select * from " + view + search + order + ";" +
                "select count(1) as items from " + table_name + " limit 1;"
            ))["data"][1]["table"][0]["items"]
        except:
            pass

        items_per_page = int(items_per_page)
        pages = -(items // -items_per_page)
        page = int(page)

        if page < 1:
            page = 1
        elif page > pages:
            page = pages

        results = self.query((
            "select * from " + table_name + " limit {from}, {to};" +
            "drop table " + table_name + ";"
        ), {"from" : (page - 1) * items_per_page, "to" : items_per_page})["data"][0]["table"] if items else []

        self.query("drop table if exists " + table_name + ";")

        return {
            "pages" : pages,
            "items" : items,
            "results" : results,
            "error" : error
        }

if "AnP" not in globals():
    class AnP:
        pass
if not hasattr(AnP, "Drivers"):
    class Drivers:
        pass
    AnP.Drivers = Drivers
    del globals()["Drivers"]
AnP.Drivers.SQLite = Anonymous
del globals()["Anonymous"]
